unit ID3;

interface
const
  ID3_OK                = 0;
  ID3_OPEN_FILE         = -1;
  ID3_READ              = -2;
  ID3_WRITE             = -3;
  ID3_NO_TAG            = -4;
  ID3_TOO_SMALL         = -5;
  ID3_OTHER             = -6;

type
  // ID3 v1.1
  ID3tag = class
    private
        data: array [0..127] of char;
        version: integer;
        fileName: string;
        loaded: boolean;
    public
      constructor Create(const filen: string);
      function ReadID3: integer;
      function SaveID3: integer;
      function RemoveID3: integer;
      procedure setVersion(const major, minor: byte);
      procedure setItem(const index: integer; const s: string);
      procedure setYear(const s: string);
      procedure setComment(const s: string);
      procedure setGenre(const g: byte);
      procedure initData;

      property getFileName: string read fileName;
      property isLoaded: boolean read loaded;
      property song: string index 3 write setItem;
      property artist: string index 33 write setItem;
      property album: string index 63 write setItem;
      property year: string write setYear;
      property comment: string write setComment;
      property genre: byte write setGenre;
    end;

implementation

uses sysutils, classes;

constructor ID3tag.Create(const filen: string);
begin
  fileName:=filen;
  loaded:=false;
  version:=1;
  initData;
end;

procedure ID3tag.initData;
begin
  data[0]:='T';
  data[1]:='A';
  data[2]:='G';
  fillchar(data[3], sizeof(data)-4, 0);
  genre:=255;
end;

function ID3tag.ReadID3: integer;
var
  f: TFileStream;
begin
  result:=ID3_OK;
  f:=nil;
  try
    f:=TFileStream.Create(fileName, fmOpenRead or fmShareDenyWrite);
    if (f.Size <= 128) then begin
      f.Free;
      result:=ID3_TOO_SMALL;
      exit;
    end;
    f.Seek(-128, soFromEnd);
    f.ReadBuffer(data, sizeof(data));
    f.Free;
    if (data[0] = 'T') and (data[1] = 'A') and (data[2] = 'G') then begin
      if (data[125] = #0) and (data[126] <> #0) then
        setVersion(1, 1);
      loaded:=true;
    end else begin
      initData;
      result:=ID3_NO_TAG;
    end;
  except
    on EFOpenError do result:=ID3_OPEN_FILE;
    on EReadError do begin
      result:=ID3_READ;
      f.Free;
    end;
  end;
end;

function ID3tag.SaveID3: integer;
var
  f: TFileStream;
  buf: array [0..2] of char;
begin
  result:=ID3_OK;
  f:=nil;
  try
    f:=TFileStream.Create(fileName, fmOpenReadWrite or fmShareExclusive);
    if (f.Size <= 128) then begin
      f.Free;
      result:=ID3_TOO_SMALL;
      exit;
    end;
    f.Seek(-128, soFromEnd);
    f.ReadBuffer(buf, 3);
    if (buf[0] = 'T') and (buf[1] = 'A') and (buf[2] = 'G') then
      f.Seek(-3, soFromCurrent)
    else
      f.Seek(0, soFromEnd);
    f.WriteBuffer(data, sizeof(data));
    loaded:=true;
    f.Free;
  except
    on EFOpenError do result:=ID3_OPEN_FILE;
    on EReadError do begin
      result:=ID3_READ;
      f.Free;
    end;
    on EWriteError do begin
      result:=ID3_WRITE;
      f.Free;
    end;
  end;
end;

function ID3tag.RemoveID3: integer;
var
  f: TFileStream;
  buf: array [0..2] of char;
begin
  result:=ID3_OK;
  f:=nil;
  try
    f:=TFileStream.Create(fileName, fmOpenReadWrite or fmShareExclusive);
    if (f.Size <= 128) then begin
      f.Free;
      result:=ID3_TOO_SMALL;
      exit;
    end;
    f.Seek(-128, soFromEnd);
    f.ReadBuffer(buf, 3);
    f.Seek(-3, soFromCurrent);
    if (buf[0] = 'T') and (buf[1] = 'A') and (buf[2] = 'G') then begin
      f.Size:=f.Size - 128;
      loaded:=false;
    end else
      result:=ID3_NO_TAG;
    f.Free;
  except
    on EFOpenError do result:=ID3_OPEN_FILE;
    on EReadError do begin
      result:=ID3_READ;
      f.Free;
    end;
    on EWriteError do begin
      result:=ID3_WRITE;
      f.Free;
    end;
    on EWin32Error do begin
      result := ID3_OTHER;
      f.Free;
    end;
  end;
end;

procedure ID3tag.setVersion(const major, minor: byte);
begin
  if (major = 1) and ((minor = 1) or (minor = 0)) then
    version:=(major shl 8) + minor;
end;

procedure ID3tag.setItem(const index: integer; const s: string);
var
  s1: string[30];
begin
  s1:=s;
  move(s1[1], data[index], length(s1));
  if (length(s1) < 30) then
    data[index+length(s1)]:=#0;
end;

procedure ID3tag.setYear(const s: string);
var
  s1: string[4];
begin
  s1:=s;
  move(s1[1], data[93], length(s1));
  if (length(s1) < 30) then
    data[93+length(s1)]:=#0;
end;

procedure ID3tag.setComment(const s: string);
var
  s1: string[30];
  max: byte;
begin
  if (version = $101) then
    max:=28
  else
    max:=30;
  s1:=copy(s, 1, max);
  move(s1[1], data[97], length(s1));
  if (length(s1) < max) then
    data[97+length(s1)]:=#0;
end;

procedure ID3tag.setGenre(const g: byte);
begin
  data[127]:=char(g);
end;

end.


